#include <stddef.h>
#include "mimetypes.h"

/* MIME types, for use in Skeleton. This must be in the same order as
 * enum OggzStreamContent defined in <oggz/oggz_constants.h> */
const char * const mime_type_names[] = {
  "video/theora",
  "audio/vorbis",
  "audio/speex",
  "audio/x-oggpcm",
  "text/cmml",
  "application/x-annodex-2",
  NULL, /* Skeleton */
  "audio/flac",
  "audio/flac",
  NULL, /* AnxData */
  "audio/celt",
  "application/x-kate",
  "video/dirac",
  "audio/x-opus",
  "video/vp8",
  NULL /* UNKNOWN */
};


