#ifndef OGGZ_MIME_TYPES_H
#define OGGZ_MIME_TYPES_H

#ifdef __cplusplus
extern "C" {
#endif

extern const char * const mime_type_names[];

#ifdef __cplusplus
}
#endif

#endif  /* OGGZ_MIME_TYPES_H */






